/* Naive solution to FIFO State Pen problem:  delete names from hash table

   Part of example code for "Drill Bits" column in ACM Queue magazine
   November/December 2023,  https://queue.acm.org/DrillBits

   Copyright (C) 2023  Terence Kelly
   Contact:  tpkelly @ { acm.org, cs.princeton.edu, eecs.umich.edu }

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU Affero General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Affero General Public License for more details.

   You should have received a copy of the GNU Affero General Public License
   along with this program.  If not, see <https://www.gnu.org/licenses/>.

   See the README file for notes on the design rationale.
*/

#include <cinttypes>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <string>
#include <unordered_set>
#include <vector>

#include "util.h"

using namespace std;

int main(int argc, char *argv[]) {
  if (3 != argc)
    BAIL("usage:  "     << argv[0] << "  capacity  escapee");
  uintmax_t capacity = s2u(argv[1]);
  string name, escapee =   argv[2];

  // ingest input strings into I
  vector<string> I;
  I.reserve(capacity);
  int flag = 0;
  uint64_t Tmethod, T0 = Tns();
  while (cin >> name) {
    I.push_back(name);
    if (escapee == name && flag++)
      BAIL("name '" << escapee << "' on input twice");
  }
  uint64_t T1 = Tns();
  if (! flag)
    BAIL("name '" << escapee << "' not on input");
  if (I.size() != capacity)
    PE("WARNING:  number of inputs " << I.size() << " != capacity");
  P("time to read input: " << T1 - T0);

  unordered_set<string> H;
  H.reserve(capacity);
  T0 = Tns();
  // fill H with all inmate names
  for (auto vi = I.begin(); vi != I.end(); ++vi)
    H.insert(*vi);
  T1 = Tns();
  P("time to fill hash table: " << T1 - T0);
  Tmethod = T1 - T0;

  T0 = Tns();
  // remove escapee name from I
  for (auto vi = I.begin(); vi != I.end(); ++vi)
    if (escapee == *vi) {
      auto last = I.rbegin();
      *vi = *last;   // overwrite escapee name with last name
      I.pop_back();  // remove last name
      break;
    }
  T1 = Tns();
  P("time to remove escapee name from I: " << T1 - T0);

  // TODO:  randomly shuffle I here (?)

  // remove all non-escaped inmates from H
  T0 = Tns();
  for (auto vi = I.begin(); vi != I.end(); ++vi)
    H.erase(*vi);
  T1 = Tns();
  P("time to empty hash table: " << T1 - T0);
  Tmethod += T1 - T0;
  P("TIME hashdel " << capacity << " " << Tmethod);

  // report the name of the escapee
  auto i = H.begin();
  P("remaining name in hash table: " << *i);
  // make sure that only one name remained in hash table
  if (++i != H.end())
    BAIL("extra remaining name: " << *i);

  return 0;
}

