/* Generate stream of unique random strings (no repetition).
   Author:  Terence Kelly
   Contact:  tpkelly @ { acm.org, cs.princeton.edu, eecs.umich.edu }
   Part of example code for "Drill Bits" column in ACM Queue magazine
   November/December 2023,  https://queue.acm.org/DrillBits
*/

#include <assert.h>
#include <errno.h>
#include <inttypes.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "util.h"

/* XORshift pseudo-random number generator from George Marsaglia's
   paper.  This is a full-period PRNG, so the same value won't be
   returned twice unless it is called 2^64 times. */
static uint64_t xor64(void) {
  static uint64_t x = UINT64_C(88172645463325252);
  x ^= (x<<13);
  x ^= (x>>7);
  return (x ^= (x<<17));
}

int main(int argc, char *argv[]) {
  uintmax_t N;
  if (2 != argc)
    BAIL("usage:  %s  N\n", argv[0]);
  N = s2u(argv[1]);
  if (0 >= N)
    BAIL("bad N '%s' -> %" PRIuMAX "\n", argv[1], N);
  while (N--) {
    char A[] = "abcdefghij",
         B[21];  // enough for base-10 UINT64_MAX + trailing nul byte
    int n = snprintf(B, sizeof B, "%020" PRIu64, xor64());
    assert(n < (int)sizeof B);  // truncation should be impossible
    for (int i = 0; B[i]; i++)
      B[i] = A[B[i] - '0'];  // like "tr 0-9 a-j"
    P("%s\n", B);
  }
  return 0;
}

